#include "stdafx.h"
#include "SearchFile.h"
#include "JsonParser.h"
#include "JsonData.h"
#include "Emule.h"
#include "EmuleDlg.h"
#include "UserMsgs.h"
#include "SearchParams.h"
#include "otherfunctions.h"
#include "log.h"
#include <afxinet.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

using namespace Json;

class Context_VC : public Context{
public:
	Context_VC(uint32 _nSearchID)
		: Context(_nSearchID)
		, inrecordarray(false)
	{
	}
	bool inrecordarray;
};

bool vc_array_beg(void* ctx, const Json::Value* value){
	if(((Context_VC*)ctx)->jp->GetLevel() == 1)
		((Context_VC*)ctx)->inrecordarray = (((Context_VC*)ctx)->key == KRecordArray);
	return true;
}

bool vc_object_end(void* ctx, const Json::Value* value){
	if(((Context_VC*)ctx)->inrecordarray && ((Context_VC*)ctx)->jp->GetLevel() == 3){
		theApp.emuledlg->SendMessage(UM_ADDSEARCHRECORD, 0, (LPARAM)ctx);
		((Context_VC*)ctx)->curRecord.reset();
	}
	return true;
}

bool vc_getkey(void* ctx, const Json::Value* value){
	if(((Context_VC*)ctx)->inrecordarray && ((Context_VC*)ctx)->jp->GetLevel() == 3){
		SearchRecord& record = ((Context_VC*)ctx)->curRecord;
		if(!strcmp(value->value, "name"))
			((Context_VC*)ctx)->key = KName;
		/*else if(!strcmp(value->value, "exts")){
		}*/
		/*else if(!strcmp(value->value, "time")){
		}*/
		else if(!strcmp(value->value, "hash"))
			((Context_VC*)ctx)->key = KHash;
		else if(!strcmp(value->value, "aich"))
			((Context_VC*)ctx)->key = KAICHHash;
		else if(!strcmp(value->value, "bytes"))
			((Context_VC*)ctx)->key = KSize;
		//else if(!strcmp(value->value, "status"))
		else if(!strcmp(value->value, "weeklyStart"))
			((Context_VC*)ctx)->key = KSource;
		else if(!strcmp(value->value, "weeklyFinish"))
			((Context_VC*)ctx)->key = KCompletedSource;
	}
	else if(((Context_VC*)ctx)->jp->GetLevel() == 1){
		if(!strcmp(value->value, "recordsReturned") || !strcmp(value->value, "start"))
			((Context_VC*)ctx)->key = KRecordMinus;
		else if(!strcmp(value->value, "total"))
			((Context_VC*)ctx)->key = KRecordPlus;
		else if(!strcmp(value->value, "records"))
			((Context_VC*)ctx)->key = KRecordArray;
	}
	return true;
}

bool vc_getvalue(void* ctx, const Json::Value* value){
	if(((Context_VC*)ctx)->key == KNone)
		return true;
	SearchRecord& record = ((Context_VC*)ctx)->curRecord;
	if(value->type == Json::T_STRING){
		switch(((Context_VC*)ctx)->key){
			case KName:
				if(!record.sName && *value->value)
					record.sName = nstrdup(value->value);
				break;
			case KHash:
				if(!record.hasICH && *value->value && strlen(value->value) == 16*2)
				{
					if(strmd4(value->value, record.sHash))
						record.hasICH = true;
				}
				break;
			case KAICHHash:
				if(!record.hasAICH && *value->value)
				{
					if(DecodeBase32(value->value, record.sAich, HASHSIZE))
						record.hasAICH = true;
				}
				break;
		}
	}
	else if(value->type == Json::T_INTEGER){
		switch(((Context_VC*)ctx)->key){
			case KSize:
				record.iSize = _strtoui64(value->value, NULL, 10);
				break;
			case KSource:
				record.iSource = strtoul(value->value, NULL/*, 10*/);
				break;
			case KCompletedSource:
				record.iCompletedSource = strtoul(value->value, NULL/*, 10*/);
				break;
			case KRecordMinus:
				((Context_VC*)ctx)->rest -= strtoul(value->value, NULL/*, 10*/);
				break;
			case KRecordPlus:
				((Context_VC*)ctx)->rest += strtoul(value->value, NULL/*, 10*/);
				break;
		}
	}
	((Context_VC*)ctx)->key = KNone;
	return true;
}

void JsonSearchThread::run()
{
	DbgSetThreadName("JsonSearch");

	//Xman
	// BEGIN SLUGFILLER: SafeHash
	CReadWriteLock lock(&theApp.m_threadlock);
	if (!lock.ReadLock(0))
		return;
	// END SLUGFILLER: SafeHash

	CInternetSession internetSession;  
	CHttpFile*pFile=NULL;
	try  
	{  
		internetSession.SetOption(INTERNET_OPTION_CONNECT_TIMEOUT, 500 );
		CString url;
		switch(pParams->eType){
		case SearchTypeVeryCD:
			url.Format(_T("http://verycd.com/search/files/%s?format=json&sort=weeklyFinish&start=%u&num=") chSTR(MAX_JSON_SEARCH_COUNT), pParams->strExpression, pParams->uStartIndex);
			break;
		default:
			ASSERT(0);
			break;
		}
		pFile = (CHttpFile*)internetSession.OpenURL(url); 
	}  
	catch(CInternetException* pEx)  
	{  
		pEx->Delete();
		return;
	}
	if(pFile){
		//#define USEBUFFER
#ifdef USEBUFFER
		char buf[512];
#else
		char in;
#endif
		Json::CJsonParser jp;
		Context* ctx = NULL;
		switch (pParams->eType){
		case SearchTypeVeryCD:{
			ctx = new Context_VC(pParams->dwSearchID);
			jp.SetContext(ctx);
			ctx->jp = &jp;
			jp.callbacks[Json::T_ARRAY_BEGIN]=&vc_array_beg;
			jp.callbacks[Json::T_OBJECT_END]=&vc_object_end;
			jp.callbacks[Json::T_KEY]=&vc_getkey;
			jp.callbacks[Json::T_VALUE]=&vc_getvalue;
							  }
							  break;
		}
#ifdef USEBUFFER
		bool parseEnd = false;
		while(pFile->Read(buf, _countof(buf))){
			for(INT_PTR i = 0; i< _countof(buf); ++i){
				if(!jp.Next(buf[i])){
					parseEnd = true;
					break;
				}
			}
			if(parseEnd)
				break;
		}
#else
		while(pFile->Read(&in, 1)){
			if(!jp.Next(in))
				break;
		}
#endif
		theApp.emuledlg->SendMessage(UM_ADDSEARCHRECORD, 1, (LPARAM)ctx);
		if (!jp.isDone())
			AddDebugLogLine(false, _T("JSON_parser_char: syntax error"));
		delete ctx;
		pFile->Close();
		delete pFile;  
		internetSession.Close(); 
	}
	else
		theApp.emuledlg->SendMessage(UM_ADDSEARCHRECORD, 2, (LPARAM)pParams->dwSearchID);
}
