#pragma once
/* Copyright (c) 2005 JSON.org */
namespace Json{
enum EType
{
	T_NONE = 0,
	T_ARRAY_BEGIN,/* event */
	T_ARRAY_END,/* event */
	T_OBJECT_BEGIN,/* event */
	T_OBJECT_END,/* event */
	T_KEY,/* event */
	T_VALUE,/* event */
	T_ITEM_END,/* event */
	T_EVENT_MAX,
	T_INTEGER,
	T_FLOAT,
	T_NULL,
	T_TRUE,
	T_FALSE,
	T_STRING
};

enum EMode;
enum EState;

struct Value {
	EType type;
	size_t length;
	char* value;
};

/*! \brief JSON parser callback 

\param ctx The pointer passed to new_parser.
\param type An element of type but not T_NONE.    
\param value A representation of the parsed value. This parameter is NULL for
T_ARRAY_BEGIN, T_ARRAY_END, T_OBJECT_BEGIN, T_OBJECT_END,
T_NULL, T_TRUE, and SON_T_FALSE. String values are always returned
as zero-terminated C strings.

\return Non-zero if parsing should continue, else zero.
*/
typedef bool (*parser_callback)(void* ctx/*, int type*/, const Value* value);
typedef unsigned short UTF16;

class CJsonParser{
public:
	/*! \brief The structure used to configure a JSON parser object 

	\param Pointer to a callback. This parameter may be NULL. In this case the input is merely checked for validity.
	*/
	parser_callback callbacks[T_EVENT_MAX];

	CJsonParser(sint_ptr _depth = 128, sint_ptr _init_buffer_capacity = 512, bool _allow_comments = true);
	void SetContext(void* _callback_ctx){callback_ctx = _callback_ctx;}

	/*! \brief Destroy a previously created JSON parser object. */
	~CJsonParser();

	/*! \brief Parse a character.

	\return Non-zero, if all characters passed to this function are part of are valid JSON.
	*/
	bool Next(unsigned int next_char);
	INT_PTR GetLevel() const{	return top - stack;	}

	/*! \brief Finalize parsing.

	Call this method once after all input characters have been consumed.

	\return Non-zero, if all parsed characters are valid JSON, zero otherwise.
	*/
	bool isDone();
	/*! \brief Determine if a given string is valid JSON white space 

	\return Non-zero if the string is valid, zero otherwise.
	*/
	//int parser_is_legal_white_space_string(const char* s);
private:
	void grow_parse_buffer();
	bool parse_parse_buffer();
	bool decode_unicode_char();
	bool add_escaped_char_to_parse_buffer(int next_char);

	/*!
	\param Callback context. This parameter may be NULL.
	\param depth. Specifies the levels of nested JSON to allow. Negative numbers yield unlimited nesting.
	\param allowComments. To allow C style comments in JSON, set to non-zero.
	\param handleFloatsManually. To decode floating point numbers manually set this parameter to non-zero
	*/
	void* callback_ctx;
	sint_ptr depth;
	bool allow_comments, comment, escaped;
	EType type;
	sint_ptr state, before_comment_state;
	UTF16 utf16_high_surrogate;
	char* parse_buffer;
	size_t parse_buffer_capacity;
	size_t parse_buffer_count;
    char* top;
    char* stack;
    char* stacktop;
	bool push(EMode mode);
	bool pop(EMode mode);
};
}