#include "StdAfx.h"
#include <afxinet.h>
#include "ModVersionCheck.h"
#include "Preferences.h"
#include "OtherFunctions.h"
#include "langids.h"
#include "ModInfo.h"
#include "Resource.h"
#include "Log.h"
#include "UserMsgs.h"
#include "Emule.h"
#include "EmuleDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

ModVersionCheck* ModVersionCheck::m_thread = NULL;

//X version check
ModVersionCheck::ModVersionCheck() : Thread(true, PRIO_LOW)
{
}

void ModVersionCheck::check() {
	if (m_thread) return;
	if (thePrefs.GetLastMVC()!=0) {
		CTime last(thePrefs.GetLastMVC());
		struct tm tmTemp;
		time_t tLast=safe_mktime(last.GetLocalTm(&tmTemp));
		time_t tNow=safe_mktime(CTime::GetCurrentTime().GetLocalTm(&tmTemp));
		if ( (difftime(tNow,tLast) / 86400)<thePrefs.GetUpdateDays() )
			return;
	}
	m_thread = new ModVersionCheck();
	if(!m_thread->start())
		ATLTRACE(_T("Failed to create version check thread!\n"));
}

void ModVersionCheck::stop()
{
	if (m_thread)
	{
		if (!m_thread->join(1000))
		{
			m_thread->terminate();
			delete m_thread;
		}
	}
}

void ModVersionCheck::run()
{
	DbgSetThreadName("ModVersionCheck");

	//Xman
	// BEGIN SLUGFILLER: SafeHash
	CReadWriteLock lock(&theApp.m_threadlock);
	if (!lock.ReadLock(0))
		return;
	// END SLUGFILLER: SafeHash

	CInternetSession internetSession;  
	CHttpFile*pFile=NULL;
	try  
	{  
		internetSession.SetOption(INTERNET_OPTION_CONNECT_TIMEOUT, 500 );
		pFile = (CHttpFile*)internetSession.OpenURL((thePrefs.GetLanguageID() == LANGID_ZH_CN)?NEW_VER_URL_CN:NEW_VER_URL); 
		if(pFile){
			pFile->Read(_buffer, _countof(_buffer) - 1);
			pFile->Close();
			delete pFile;  
			internetSession.Close(); 
			_buffer[_countof(_buffer) - 1] = 0;
			if(theApp.emuledlg)
			{
				theApp.emuledlg->PostMessage(UM_XVERSIONCHECK_RESPONSE, 0, 0);
				_deleteOnExit = false;
			}
			else
				m_thread = NULL;
			thePrefs.UpdateLastMVC();
		}
	}  
	catch(CInternetException*   pEx)  
	{
		pEx->Delete();
		AddLogLine(true,GetResString(IDS_NEWVERSIONFAILED));
		m_thread = NULL;
	}
}

const char* ModVersionCheck::getMessage()
{
	if (!m_thread) return NULL;
	const char*buffer = m_thread->_buffer;
	const char*pnext;
	if(((strtoul(buffer,(const char**)&pnext/*,10*/)>MAIN_VER_CODE) || (pnext != NULL && strtoul(pnext,(const char**)&pnext/*,10*/)>MIN_VER_CODE))
		&& (pnext=strchr(pnext, '\n')) != NULL)
			return pnext+1;
	m_thread = NULL;
	return NULL;
}
